# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
from os.path import join, isfile
import pypdfium2 as pdfium
from ..conftest import TestFiles, OutputDir


def test_save():
    
    src_pdf = pdfium.PdfDocument(TestFiles.multipage)
    new_pdf_raw = pdfium.FPDF_ImportNPagesToOne(
        src_pdf.raw,
        595, 842,
        2, 2,
    )
    
    new_pdf = pdfium.PdfDocument(new_pdf_raw)
    assert len(new_pdf) == 1
    page = new_pdf.get_page(0)
    assert page.get_size() == (595, 842)
    
    output_file = join(OutputDir, "tiling.pdf")
    with open(output_file, "wb") as buffer:
        new_pdf.save(buffer)
    assert isfile(output_file)
    
    [g.close() for g in (page, new_pdf, src_pdf)]


def test_save_withversion():
    
    pdf = pdfium.PdfDocument(TestFiles.multipage)
    pdf.del_page(1)
    
    buffer = io.BytesIO()
    pdf.save(buffer, version=17)
    pdf.close()
    
    buffer.seek(0)
    data = buffer.read()
    buffer.seek(0)
    
    exp_start = b"%PDF-1.7"
    exp_end = b"%EOF\r\n"
    assert data[:len(exp_start)] == exp_start
    assert data[-len(exp_end):] == exp_end
    
    reopened_pdf = pdfium.PdfDocument(buffer)
    assert len(reopened_pdf) == 2
    reopened_pdf.close()
    buffer.close()
    



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import pytest
import logging
import contextlib
import pypdfium2 as pdfium
from ..conftest import TestFiles


@pytest.fixture
def toc_doc():
    doc = pdfium.PdfDocument(TestFiles.toc)
    yield doc
    doc.close()

@pytest.fixture
def toc_circular_doc():
    doc = pdfium.PdfDocument(TestFiles.toc_circular)
    yield doc
    doc.close()


def _compare_bookmark(bookmark, title, page_index, view_mode, view_pos, is_closed):
    assert bookmark.title == title
    assert bookmark.page_index == page_index
    assert bookmark.view_mode == view_mode
    assert pytest.approx(bookmark.view_pos, abs=1) == view_pos
    assert bookmark.is_closed == is_closed


def test_gettoc(toc_doc):
    
    toc = toc_doc.get_toc()
    
    # check first bookmark
    _compare_bookmark(
        next(toc),
        title = "One",
        page_index = 0,
        view_mode = pdfium.PDFDEST_VIEW_XYZ,
        view_pos = (89, 758, 0),
        is_closed = True,
    )
    
    # check common values
    for bookmark in toc:
        assert isinstance(bookmark, pdfium.OutlineItem)
        assert bookmark.view_mode is pdfium.PDFDEST_VIEW_XYZ
        assert round(bookmark.view_pos[0]) == 89
    
    # check last bookmark
    _compare_bookmark(
        bookmark,
        title = "Three-B",
        page_index = 1,
        view_mode = pdfium.PDFDEST_VIEW_XYZ,
        view_pos = (89, 657, 0),
        is_closed = False,
    )


def test_gettoc_circular(toc_circular_doc, caplog):
    toc = toc_circular_doc.get_toc()
    _compare_bookmark(
        next(toc),
        title = "A Good Beginning",
        page_index = -1,
        view_mode = pdfium.PDFDEST_VIEW_UNKNOWN_MODE,
        view_pos = [],
        is_closed = False,
    )
    _compare_bookmark(
        next(toc),
        title = "A Good Ending",
        page_index = -1,
        view_mode = pdfium.PDFDEST_VIEW_UNKNOWN_MODE,
        view_pos = [],
        is_closed = False,
    )
    with caplog.at_level(logging.WARNING):
        for other in toc: pass
    assert "circular bookmark reference" in caplog.text


def _test_printtoc(doc, exp_result, max_depth=15):
    buffer = io.StringIO()
    with contextlib.redirect_stdout(buffer):
        doc.print_toc( doc.get_toc(max_depth=max_depth) )
    assert buffer.getvalue() == exp_result


def test_toc_misc(toc_doc, toc_circular_doc):
    
    _test_printtoc(toc_doc, TocResult_A)
    _test_printtoc(toc_circular_doc, TocResult_B)
    
    viewmodes_doc = pdfium.PdfDocument(TestFiles.toc_viewmodes)
    _test_printtoc(viewmodes_doc, TocResult_C)
    viewmodes_doc.close()
    
    maxdepth_doc = pdfium.PdfDocument(TestFiles.toc_maxdepth)
    _test_printtoc(maxdepth_doc, TocResult_D, max_depth=10)
    maxdepth_doc.close()


TocResult_A = """\
[-] One -> 1  # XYZ [89.29, 757.7, 0.0]
    [+] One-A -> 1  # XYZ [89.29, 706.86, 0.0]
    [-] One-B -> 1  # XYZ [89.29, 657.03, 0.0]
        [+] One-B-I -> 1  # XYZ [89.29, 607.2, 0.0]
        [+] One-B-II -> 1  # XYZ [89.29, 557.76, 0.0]
[+] Two -> 1  # XYZ [89.29, 507.16, 0.0]
[-] Three -> 2  # XYZ [89.29, 757.7, 0.0]
    [+] Three-A -> 2  # XYZ [89.29, 706.98, 0.0]
    [+] Three-B -> 2  # XYZ [89.29, 657.15, 0.0]
"""

TocResult_B = """\
[+] A Good Beginning -> 0  # Unknown []
[+] A Good Ending -> 0  # Unknown []
"""

TocResult_C = """\
[+] XYZ 2.1-Page1 red bold -> 1  # XYZ [100.0, 100.0, 2.1]
[+] Fit-Page2 green Italic -> 2  # Fit []
[+] FitB Italic&Bold b -> 3  # FitB []
[+] FitV Page4 -> 4  # FitV [500.0]
[+] FitH Page5 -> 5  # FitH [600.0]
[+] FitR Page6 -> 6  # FitR [100.0, 100.0, 200.0, 200.0]
[+] FitBH Page7 -> 7  # FitBH [100.0]
[+] FitBV Page8 -> 8  # FitBV [100.0]
"""

TocResult_D = """\
[+] 1.outline -> 1  # FitH [746.44]
    [+] 1.1.outline -> 1  # FitH [700.88]
        [+] 1.1.1.outline -> 1  # FitH [632.54]
            [+] 1.1.1.1.outline -> 1  # FitH [632.95]
                [+] 1.1.1.1.1.outline -> 1  # FitH [597.3]
                    [+] 1.1.1.1.1.1outline -> 1  # FitH [632.95]
                        [+] 1.1.1.1.1.1.outline -> 1  # FitH [632.95]
                            [+] 1.1.1.1.1.1.1.outline -> 1  # FitH [632.95]
                                [+] 1.1.1.1.1.1.1.1.outline -> 1  # FitH [632.95]
                                    [+] 1.1.1.1.1.1.1.1.1.outline -> 1  # FitH [632.95]
[+] 2.outline -> 2  # FitH [749.48]
    [+] 2.1.outline -> 2  # FitH [699.36]
        [+] 2.1.1.outline -> 2  # FitH [628.74]
            [+] 2.1.1.1.outline -> 2  # FitH [583.18]
    [+] 2.2 outline -> 2  # FitH [515.22]
"""



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-FileCopyrightText: 2022 Anurag Bansal <anurag.bansal.585@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import math
import ctypes
from ctypes import c_float, byref
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.textpage import PdfTextPage
from pypdfium2._helpers._utils import (
    colour_tohex,
    get_colourformat,
    ColourMapping,
    RotationToConst,
    RotationToDegrees,
)
from pypdfium2._helpers.misc import (
    raise_error,
    OptimiseMode,
)

try:
    import uharfbuzz as harfbuzz
except ImportError:
    have_harfbuzz = False
else:
    have_harfbuzz = True

try:
    import PIL.Image
except ImportError:
    have_pil = False
else:
    have_pil = True


class PdfPage:
    """ Page helper class. """
    
    def __init__(self, page, pdf):
        self._page = page
        self._pdf = pdf
    
    @property
    def raw(self):
        """ FPDF_PAGE: The raw PDFium page object handle. """
        return self._page
    
    @property
    def pdf(self):
        """ PdfDocument: The document this page belongs to. """
        return self._pdf
    
    def close(self):
        """
        Close the page to release allocated memory.
        This function shall be called when finished working with the object.
        """
        pdfium.FPDF_ClosePage(self._page)
    
    
    def get_width(self):
        """
        Returns:
            float: Page width (horizontal size).
        """
        return pdfium.FPDF_GetPageWidthF(self._page)
    
    def get_height(self):
        """
        Returns:
            float: Page height (vertical size).
        """
        return pdfium.FPDF_GetPageHeightF(self._page)
    
    def get_size(self):
        """
        Returns:
            (float, float): Page width and height.
        """
        return (self.get_width(), self.get_height())
    
    def get_rotation(self):
        """
        Returns:
            int: Clockwise page rotation in degrees.
        """
        return RotationToDegrees[ pdfium.FPDFPage_GetRotation(self._page) ]
    
    def set_rotation(self, rotation):
        """ Define the absolute, clockwise page rotation (0, 90, 180, or 270 degrees). """
        pdfium.FPDFPage_SetRotation(self._page, RotationToConst[rotation])
    
    
    def _get_box(self, box_func, fallback_func):
        left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
        ret_code = box_func(self._page, byref(left), byref(bottom), byref(right), byref(top))
        if not ret_code:
            return fallback_func()
        return (left.value, bottom.value, right.value, top.value)
    
    def _set_box(self, box_func, l, b, r, t):
        if not all(isinstance(val, (int, float)) for val in (l, b, r, t)):
            raise ValueError("Box values must be int or float.")
        box_func(self._page, l, b, r, t)
    
    def get_mediabox(self):
        """
        Returns:
            (float, float, float, float):
            The page MediaBox in PDF canvas units, consisting of four coordinates (usually x0, y0, x1, y1).
            Falls back to ANSI A (0, 0, 612, 792) in case MediaBox is not defined.
        """
        return self._get_box(pdfium.FPDFPage_GetMediaBox, lambda: (0, 0, 612, 792))
    
    def set_mediabox(self, l, b, r, t):
        """
        Set the page's MediaBox by passing four :class:`float` coordinates (usually x0, y0, x1, y1).
        """
        self._set_box(pdfium.FPDFPage_SetMediaBox, l, b, r, t)
    
    def get_cropbox(self):
        """
        Returns:
            The page's CropBox (If not defined, falls back to MediaBox).
        """
        return self._get_box(pdfium.FPDFPage_GetCropBox, self.get_mediabox)
    
    def set_cropbox(self, l, b, r, t):
        """ Set the page's CropBox. """
        self._set_box(pdfium.FPDFPage_SetCropBox, l, b, r, t)
    
    def get_bleedbox(self):
        """
        Returns:
            The page's BleedBox (If not defined, falls back to CropBox).
        """
        return self._get_box(pdfium.FPDFPage_GetBleedBox, self.get_cropbox)
    
    def set_bleedbox(self, l, b, r, t):
        """ Set the page's BleedBox. """
        self._set_box(pdfium.FPDFPage_SetBleedBox, l, b, r, t)
    
    def get_trimbox(self):
        """
        Returns:
            The page's TrimBox (If not defined, falls back to CropBox).
        """
        return self._get_box(pdfium.FPDFPage_GetTrimBox, self.get_cropbox)
    
    def set_trimbox(self, l, b, r, t):
        """ Set the page's TrimBox. """
        self._set_box(pdfium.FPDFPage_SetTrimBox, l, b, r, t)
    
    def get_artbox(self):
        """
        Returns:
            The page's ArtBox (If not defined, falls back to CropBox).
        """
        return self._get_box(pdfium.FPDFPage_GetArtBox, self.get_cropbox)
    
    def set_artbox(self, l, b, r, t):
        """ Set the page's ArtBox. """
        self._set_box(pdfium.FPDFPage_SetArtBox, l, b, r, t)
    
    
    def get_textpage(self):
        """
        Returns:
            PdfTextPage: The text page that corresponds to this page.
        """
        textpage = pdfium.FPDFText_LoadPage(self._page)
        if not textpage:
            raise_error("Loading the text page failed")
        return PdfTextPage(textpage, self)
    
    
    def insert_text(
            self,
            text,
            pos_x,
            pos_y,
            font_size,
            hb_font,
            pdf_font,
        ):
        """
        *Requires* uharfbuzz.
        
        Insert text into the page at a specified position, using the writing system's ligature.
        This function supports Asian scripts such as Hindi.
        
        Parameters:
            text (str):
                The message to insert.
            pos_x (float):
                Distance from left border to first character.
            pos_y (float):
                Distance from bottom border to first character.
            font_size (float):
                Font size the text shall have.
            hb_font (HarfbuzzFont):
                Harfbuzz font data.
            pdf_font (PdfFont):
                PDF font data.
        """
        
        hb_buffer = harfbuzz.Buffer()
        hb_buffer.add_str(text)
        hb_buffer.guess_segment_properties()
        hb_features = {"kern": True, "liga": True}
        harfbuzz.shape(hb_font.font, hb_buffer, hb_features)
        
        start_point = pos_x
        for info, pos in zip(hb_buffer.glyph_infos, hb_buffer.glyph_positions):
            pdf_textobj = pdfium.FPDFPageObj_CreateTextObj(self._pdf.raw, pdf_font.raw, font_size)
            pdfium.FPDFText_SetCharcodes(pdf_textobj, ctypes.c_uint32(info.codepoint), 1)
            pdfium.FPDFPageObj_Transform(
                pdf_textobj,
                1, 0, 0, 1,
                start_point - (pos.x_offset / hb_font.scale) * font_size,
                pos_y,
            )
            pdfium.FPDFPage_InsertObject(self._page, pdf_textobj)
            start_point += (pos.x_advance / hb_font.scale) * font_size
        
        pdfium.FPDFPage_GenerateContent(self._page)
    
    
    def count_objects(self):
        """
        Returns:
            int: The number of page objects on this page.
        """
        return pdfium.FPDFPage_CountObjects(self._page)
    
    def get_objects(self):
        """
        Iterate through the page objects on this page.
        
        Yields:
            :class:`.PdfPageObject`: Page object helper.
        """
        for i in range( self.count_objects() ):
            raw_obj = pdfium.FPDFPage_GetObject(self._page, i)
            yield PdfPageObject(raw_obj)
    
    
    def render_base(
            self,
            scale = 1,
            rotation = 0,
            crop = (0, 0, 0, 0),
            colour = (255, 255, 255, 255),
            annotations = True,
            greyscale = False,
            optimise_mode = OptimiseMode.NONE,
            no_antialias = (),
        ):
        """
        Rasterise the page to a ctypes ubyte array.
        This is the base function for :func:`.render_tobytes` and :func:`.render_topil`.
        
        Parameters:
            
            scale (float):
                Define the quality (or size) of the image.
                By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the number of pixels that represent one point.
                Higher values increase quality, file size and rendering duration, while lower values reduce them.
                Note that UserUnit is not taken into account, so if you are using pypdfium2 in conjunction with an other PDF library, you may want to check for a possible ``/UserUnit`` in the page dictionary and multiply this scale factor with it.
            
            rotation (int):
                A rotation value in degrees (0, 90, 180, or 270), in addition to page rotation.
            
            crop (typing.Tuple[float, float, float, float]):
                Amount in PDF canvas units to cut off from page borders (left, bottom, right, top).
                Crop is applied after rotation.
            
            colour (typing.Optional[ typing.Tuple[int, int, int, typing.Optional[int]] ]):
                Page background colour. Defaults to white.
                It can either be :data:`None`, or values of red, green, blue, and alpha ranging from 0 to 255.
                If :data:`None`, the bitmap will not be filled with a colour, resulting in transparent background.
                For RGB, 0 will include nothing of the colour in question, while 255 will completely include it. For Alpha, 0 means full transparency, while 255 means no transparency.
            
            annotations (bool):
                Whether to render page annotations.
            
            greyscale (bool):
                Whether to render in greyscale mode (no colours).
            
            optimise_mode (OptimiseMode):
                How to optimise page rendering.
            
            no_antialias (typing.Sequence[str]):
                A list of item types that shall not be smoothed (text, image, path).
        
        Returns:
            (BitmapDataHolder, str, (int, int)):
            Bitmap data holder, colour format, and image size.
            The colour format can be ``BGRA``, ``BGR``, or ``L``, depending on the parameters *colour* and *greyscale*.
            Image size is given in pixels as a tuple of width and height.
        """
        
        if colour is None:
            fpdf_colour, use_alpha = None, True
        else:
            fpdf_colour, use_alpha = colour_tohex(*colour)
        
        cl_format, cl_pdfium = get_colourformat(use_alpha, greyscale)
        n_colours = len(cl_format)
        
        form_config = pdfium.FPDF_FORMFILLINFO(2)
        form_fill = pdfium.FPDFDOC_InitFormFillEnvironment(self._pdf.raw, form_config)
        
        src_width  = math.ceil(self.get_width()  * scale)
        src_height = math.ceil(self.get_height() * scale)
        if rotation in (90, 270):
            src_width, src_height = src_height, src_width
        
        crop = [math.ceil(c*scale) for c in crop]
        width  = src_width  - crop[0] - crop[2]
        height = src_height - crop[1] - crop[3]
        if any(d < 1 for d in (width, height)):
            raise ValueError("Crop exceeds page dimensions (in px): width %s, height %s, crop %s" % (src_width, src_height, crop))
        
        bitmap = pdfium.FPDFBitmap_CreateEx(width, height, cl_pdfium, None, width*n_colours)
        if fpdf_colour is not None:
            pdfium.FPDFBitmap_FillRect(bitmap, 0, 0, width, height, fpdf_colour)
        
        render_flags = 0
        
        if annotations:
            render_flags |= pdfium.FPDF_ANNOT
        if greyscale:
            render_flags |= pdfium.FPDF_GRAYSCALE
        
        if "text" in no_antialias:
            render_flags |= pdfium.FPDF_RENDER_NO_SMOOTHTEXT
        if "image" in no_antialias:
            render_flags |= pdfium.FPDF_RENDER_NO_SMOOTHIMAGE
        if "path" in no_antialias:
            render_flags |= pdfium.FPDF_RENDER_NO_SMOOTHPATH
        
        if optimise_mode is OptimiseMode.NONE:
            pass
        elif optimise_mode is OptimiseMode.LCD_DISPLAY:
            render_flags |= pdfium.FPDF_LCD_TEXT
        elif optimise_mode is OptimiseMode.PRINTING:
            render_flags |= pdfium.FPDF_PRINTING
        else:
            raise ValueError("Invalid optimise_mode %s" % optimise_mode)
        
        render_args = (bitmap, self._page, -crop[0], -crop[3], src_width, src_height, RotationToConst[rotation], render_flags)
        pdfium.FPDF_RenderPageBitmap(*render_args)
        pdfium.FPDF_FFLDraw(form_fill, *render_args)
        
        cbuf_ptr = pdfium.FPDFBitmap_GetBuffer(bitmap)
        cbuf_array_ptr = ctypes.cast(cbuf_ptr, ctypes.POINTER(ctypes.c_ubyte * (width*height*n_colours)))
        data_holder = BitmapDataHolder(bitmap, cbuf_array_ptr)
        
        pdfium.FPDFDOC_ExitFormFillEnvironment(form_fill)
        
        return data_holder, cl_format, (width, height)
    
    
    def render_tobytes(self, *args, **kwargs):
        """
        Rasterise the page to bytes. Parameters are the same as for :meth:`.render_base`.
        
        Returns:
            (bytes, str, (int, int)): Image data, colour format, and size.
        """
        
        data_holder, cl_format, size = self.render_base(*args, **kwargs)
        data = bytes( data_holder.get_data() )
        data_holder.close()
        
        return data, cl_format, size
    
    
    def render_topil(self, *args, **kwargs):
        """
        *Requires* :mod:`PIL`.
        
        Rasterise the page to a PIL image. Parameters are the same as for :meth:`.render_base`.
        
        Returns:
            PIL.Image.Image: An image of the page.
        """
        
        if not have_pil:
            raise RuntimeError("Pillow library needs to be installed for render_topil().")
        
        data_holder, cl_format, size = self.render_base(*args, **kwargs)
        pil_image = PIL.Image.frombytes(ColourMapping[cl_format], size, data_holder.get_data(), "raw", cl_format, 0, 1)
        data_holder.close()
        
        return pil_image


class BitmapDataHolder:
    
    def __init__(self, bm_handle, bm_array_ptr):
        self.bm_handle = bm_handle
        self.bm_array_ptr = bm_array_ptr
    
    def get_data(self):
        return self.bm_array_ptr.contents
    
    def close(self):
        pdfium.FPDFBitmap_Destroy(self.bm_handle)


class PdfPageObject:
    """ Page object helper class. """
    
    def __init__(self, pageobj):
        self._pageobj = pageobj
    
    def get_pos(self):
        """
        Get the position of the object on the page.
        
        Returns:
            A tuple of four :class:`float` coordinates for left, bottom, right, and top.
        """
        left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
        ret_code = pdfium.FPDFPageObj_GetBounds(self._pageobj, byref(left), byref(bottom), byref(right), byref(top))
        if not ret_code:
            raise_error("Locating the page object failed")
        return (left.value, bottom.value, right.value, top.value)
    
    def get_type(self):
        """
        Returns:
            int: The type of the object (``FPDF_PAGEOBJ_...``).
        """
        return pdfium.FPDFPageObj_GetType(self._pageobj)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pypdfium2._pypdfium as pdfium


def colour_tohex(r, g, b, a=255):
    """
    Convert an RGBA colour specified by four integers ranging from 0 to 255 to a single ARGB32 value.
    
    Returns:
        (int, bool): The colour value, and a boolean signifying if an alpha channel is needed.
    """
    
    use_alpha = True
    if a == 255:
        use_alpha = False
    
    colours = (a, r, g, b)
    for col in colours:
        assert 0 <= col <= 255
    
    hex_str = "0x"
    for col in colours:
        hex_str += hex(col)[2:].zfill(2)
    
    hex_int = int(hex_str, 0)
    
    return hex_int, use_alpha


def get_colourformat(use_alpha, greyscale):
    """
    Get the required colour format according to the boolean values of *use_alpha* and *greyscale*.
    
    Returns:
        (str, int): The colour format as string, and as PDFium constant (``FPDFBitmap_...``).
    """
    
    px = "BGRA", pdfium.FPDFBitmap_BGRA
    
    if not use_alpha:
        if greyscale:
            px = "L", pdfium.FPDFBitmap_Gray
        else:
            px = "BGR", pdfium.FPDFBitmap_BGR
    
    return px


def _invert_dict(dictionary):
    """
    Returns:
        A copy of *dictionary*, with inverted keys and values.
    """
    return {v: k for k, v in dictionary.items()}


#: Convert an input pixel format to its :mod:`PIL` target format.
ColourMapping = {
    "BGRA": "RGBA",
    "BGR":  "RGB",
    "L":    "L",
}

#: Convert a PDFium view mode constant (``PDFDEST_VIEW_...``) to string.
ViewmodeMapping = {
    pdfium.PDFDEST_VIEW_XYZ:   "XYZ",
    pdfium.PDFDEST_VIEW_FIT:   "Fit",
    pdfium.PDFDEST_VIEW_FITH:  "FitH",
    pdfium.PDFDEST_VIEW_FITV:  "FitV",
    pdfium.PDFDEST_VIEW_FITR:  "FitR",
    pdfium.PDFDEST_VIEW_FITB:  "FitB",
    pdfium.PDFDEST_VIEW_FITBH: "FitBH",
    pdfium.PDFDEST_VIEW_FITBV: "FitBV",
    pdfium.PDFDEST_VIEW_UNKNOWN_MODE: "Unknown",
}

#: Convert a PDFium error constant (``FPDF_ERR_...``) to string.
ErrorMapping = {
    pdfium.FPDF_ERR_SUCCESS:  "Success",
    pdfium.FPDF_ERR_UNKNOWN:  "Unknown error",
    pdfium.FPDF_ERR_FILE:     "File access error",
    pdfium.FPDF_ERR_FORMAT:   "Data format error",
    pdfium.FPDF_ERR_PASSWORD: "Incorrect password error",
    pdfium.FPDF_ERR_SECURITY: "Unsupported security scheme error",
    pdfium.FPDF_ERR_PAGE:     "Page not found or content error",
}

#: Convert a PDFium object type constant (``FPDF_PAGEOBJ_...``) to string.
ObjtypeToName = {
    pdfium.FPDF_PAGEOBJ_UNKNOWN: "unknown",
    pdfium.FPDF_PAGEOBJ_TEXT:    "text",
    pdfium.FPDF_PAGEOBJ_PATH:    "path",
    pdfium.FPDF_PAGEOBJ_IMAGE:   "image",
    pdfium.FPDF_PAGEOBJ_SHADING: "shading",
    pdfium.FPDF_PAGEOBJ_FORM:    "form",
}

#: Convert an object type string to a PDFium constant. Inversion of :data:`ObjtypeToName`.
ObjtypeToConst = _invert_dict(ObjtypeToName)

#: Convert a rotation value in degrees to a PDFium constant.
RotationToConst = {
    0:   0,
    90:  1,
    180: 2,
    270: 3,
}

#: Convert a PDFium rotation constant to a value in degrees. Inversion of :data:`RotationToConst`.
RotationToDegrees = _invert_dict(RotationToConst)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: CC-BY-4.0

# Configuration file for the Sphinx documentation builder.
# See https://www.sphinx-doc.org/en/master/usage/configuration.html

import time

project = "pypdfium2"
author = "pypdfium2-team"
copyright = "%s %s" % (time.strftime("%Y"), author)

html_theme = "sphinx_rtd_theme"
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.programoutput",
    "myst_parser",
]

add_module_names = False
autodoc_preserve_defaults = True
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "member-order": "bysource",
}
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "PIL": ("https://pillow.readthedocs.io/en/stable/", None),
}



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os.path
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parsers import pagetext_type


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "extract-text",
        help = "Extract text from a PDF page in given boundaries",
    )
    parser.add_argument(
        "input",
        type = os.path.abspath,
        help = "Path to the PDF document to work with",
    )
    parser.add_argument(
        "--password",
        help = "Password to unlock the PDF, if encrypted",
    )
    parser.add_argument(
        "--pages",
        help = "Page numbers to include (defaults to all)",
        type = pagetext_type,
    )


def main(args):
    
    doc = pdfium.PdfDocument(args.input, args.password)
    if args.pages is None:
        args.pages = [i for i in range(len(doc))]
    
    sep = ""
    for index in args.pages:
        
        page = doc.get_page(index)
        textpage = page.get_textpage()
        text = textpage.get_text()
        
        [g.close() for g in (textpage, page)]
        
        print(sep + "# Page %s\n" % (index+1) + text)
        sep = "\n"
    
    doc.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import re
import pytest
from os.path import join
from importlib.util import find_spec
import pypdfium2 as pdfium
from ..conftest import TestFiles, ResourceDir, OutputDir


@pytest.fixture
def doc():
    doc = pdfium.PdfDocument(TestFiles.text)
    yield doc
    doc.close()


@pytest.fixture
def textpage(doc):
    page = doc.get_page(0)
    textpage = page.get_textpage()
    assert isinstance(textpage, pdfium.PdfTextPage)
    yield textpage
    [g.close() for g in (textpage, page)]


@pytest.fixture
def linkpage(doc):
    page = doc.get_page(1)
    linkpage = page.get_textpage()
    yield linkpage
    [g.close() for g in (linkpage, page)]


def test_gettext(textpage):
    text = textpage.get_text()
    assert len(text) == 438
    assert text.startswith("Lorem ipsum dolor sit amet,\r\n")
    assert text.endswith("\r\nofficia deserunt mollit anim id est laborum.")


@pytest.mark.parametrize("loose", [False, True])
def test_getcharbox(textpage, loose):
    for index in range( textpage.count_chars() ):
        box = textpage.get_charbox(index, loose=loose)
        assert all( isinstance(val, (int, float)) for val in box )
        assert box[0] <= box[2] and box[1] <= box[3]


def test_getrectboxes(textpage):
    rects = textpage.get_rectboxes()
    
    first_rect = next(rects)
    assert pytest.approx(first_rect, abs=1) == [58, 767, 258, 782]
    first_text = textpage.get_text(*first_rect)
    assert first_text == "Lorem ipsum dolor sit amet,"
    
    i = 0
    for rect in rects:
        assert len(rect) == 4
        assert 56 < rect[0] < 59
        text = textpage.get_text(*rect)
        assert isinstance(text, str)
        assert len(text) <= 66
        i += 1
    
    assert i == 9
    assert text == "officia deserunt mollit anim id est laborum."


def test_gettext_area_oob(textpage):
    areas = [
        (-1, 0, 0, 0),   # left too low
        (0, -1, 0, 0),   # bottom too low
        (0, 0, -1, 0),   # right too low
        (0, 0, 0, -1),   # top too low
        (600, 0, 0, 0),  # left too high
        (0, 900, 0, 0),  # bottom too high
        (0, 0, 600, 0),  # right too high
        (0, 0, 0, 900),  # top too high
        (100, 0, 50, 0), # left higher than right
        (0, 100, 0, 50), # bottom higher than top
    ]
    for area in areas:
        with pytest.raises(ValueError, match=re.escape("Invalid page area requested.")):
            textpage.get_text(*area)


def test_search_text(textpage):
    searcher = textpage.search("labor")
    
    occ_1a = searcher.get_next()
    occ_2a = searcher.get_next()
    occ_3a = searcher.get_next()
    occ_4x = searcher.get_next()
    occ_2b = searcher.get_prev()
    occ_1b = searcher.get_prev()
    
    assert pytest.approx(occ_1a[0], abs=1) == (292, 678, 329, 690)
    assert pytest.approx(occ_2a[0], abs=1) == (324, 641, 360, 653)
    assert pytest.approx(occ_3a[0], abs=1) == (305, 549, 341, 561)
    assert occ_4x is None
    assert occ_1a == occ_1b and occ_2a == occ_2b
    
    for occ in (occ_1a, occ_2a, occ_3a):
        for box in occ:
            assert all(isinstance(val, (int, float)) for val in box)
            assert len(box) == 4
            assert box[0] <= box[2]
            assert box[1] <= box[3]
    
    searcher.close()


def test_get_index(textpage):
    
    x, y = (60, textpage.page.get_height()-66)
    
    n_chars = textpage.count_chars()
    index = textpage.get_index(x, y, 5, 5)
    assert index < n_chars and index == 0
    
    charbox = textpage.get_charbox(index)
    char = textpage.get_text(*charbox)
    assert char == "L"


def test_textpage_empty():
    pdf = pdfium.PdfDocument(TestFiles.empty)
    page = pdf.get_page(0)
    textpage = page.get_textpage()
    
    assert textpage.get_text() == ""
    assert textpage.count_chars() == 0
    assert textpage.count_rects() == 0
    assert textpage.get_index(0, 0, 0, 0) is None
    assert [r for r in textpage.get_rectboxes()] == []
    
    searcher = textpage.search("a")
    assert searcher.get_next() is None
    searcher.close()
    
    with pytest.raises(ValueError, match=re.escape("Character index 0 is out of bounds. The maximum index is -1.")):
        textpage.get_charbox(0)
    with pytest.raises(ValueError, match=re.escape("Text length must be >0.")):
        textpage.search("")
    
    [g.close() for g in (textpage, page, pdf)]


def test_get_links(linkpage):
    exp_links = (
        "https://www.wikipedia.org/",
        "https://www.openstreetmap.org/",
        "https://www.opensuse.org/",
        "https://kde.org/",
    )
    for i, link in enumerate(linkpage.get_links()):
        assert link == exp_links[i]


@pytest.mark.skipif(not find_spec("uharfbuzz"), reason="uharfbuzz is not installed")
def test_insert_text():
    
    pdf = pdfium.PdfDocument.new()
    width, height = 500, 300
    page = pdf.new_page(width, height)
    assert page.get_size() == (width, height)
    
    NotoSans = join(ResourceDir, "NotoSans-Regular.ttf")
    hb_font = pdfium.HarfbuzzFont(NotoSans)
    pdf_font = pdf.add_font(
        NotoSans,
        type = pdfium.FPDF_FONT_TRUETYPE,
        is_cid = True,
    )
    
    message_a = "मैं घोषणा, पुष्टि और सहमत हूँ कि:"
    message_b = "Latin letters test."
    page.insert_text(
        text = message_a,
        pos_x = 50,
        pos_y = height-75,
        font_size = 25,
        hb_font = hb_font,
        pdf_font = pdf_font,
    )
    page.insert_text(
        text = message_b,
        pos_x = 50,
        pos_y = height-150,
        font_size = 30,
        hb_font = hb_font,
        pdf_font = pdf_font,
    )
    
    # NOTE Extraction of hindi text currently fails (PDFium issue, confirmed with Chromium)
    textpage = page.get_textpage()
    assert textpage.get_text(left=50, bottom=height-150, top=height-120) == message_b
    
    with open(join(OutputDir, "text_insertion.pdf"), "wb") as buffer:
        pdf.save(buffer, version=17)
    
    [g.close() for g in (textpage, page, pdf_font, pdf)]
